package envoy.client.data;

import java.util.*;

import envoy.data.IdGenerator;
import envoy.data.Message;
import envoy.data.User;
import envoy.event.MessageStatusChangeEvent;

/**
 * Stores information about the current {@link User} and their {@link Chat}s.
 * For message ID generation a {@link IdGenerator} is stored as well.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>LocalDb.java</strong><br>
 * Created: <strong>3 Feb 2020</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @since Envoy v0.3-alpha
 */
public abstract class LocalDb {

	protected User								user;
	protected Map<String, User>					users			= new HashMap<>();
	protected List<Chat>						chats			= new ArrayList<>();
	protected IdGenerator						idGenerator;
	protected Cache<Message>					messageCache	= new Cache<>();
	protected Cache<MessageStatusChangeEvent>	statusCache		= new Cache<>();

	/**
	 * Initializes a storage space for a user-specific list of chats.
	 *
	 * @since Envoy v0.3-alpha
	 */
	public void initializeUserStorage() {}

	/**
	 * Stores all users. If the client user is specified, their chats will be stored
	 * as well. The message id generator will also be saved if present.
	 *
	 * @throws Exception if the saving process failed
	 * @since Envoy v0.3-alpha
	 */
	public void save() throws Exception {}

	/**
	 * Loads all user data.
	 *
	 * @throws Exception if the loading process failed
	 * @since Envoy v0.3-alpha
	 */
	public void loadUsers() throws Exception {}

	/**
	 * Loads all data of the client user.
	 *
	 * @throws Exception if the loading process failed
	 * @since Envoy v0.3-alpha
	 */
	public void loadUserData() throws Exception {}

	/**
	 * Loads the ID generator. Any exception thrown during this process is ignored.
	 *
	 * @since Envoy v0.3-alpha
	 */
	public void loadIdGenerator() {}

	/**
	 * @return a {@code Map<String, User>} of all users stored locally with their
	 *         user names as keys
	 * @since Envoy v0.2-alpha
	 */
	public Map<String, User> getUsers() { return users; }

	/**
	 * @param users the users to set
	 */
	public void setUsers(Map<String, User> users) { this.users = users; }

	/**
	 * @return all saved {@link Chat} objects that list the client user as the
	 *         sender
	 * @since Envoy v0.1-alpha
	 **/
	public List<Chat> getChats() { return chats; }

	/**
	 * @param chats the chats to set
	 */
	public void setChats(List<Chat> chats) { this.chats = chats; }

	/**
	 * @return the {@link User} who initialized the local database
	 * @since Envoy v0.2-alpha
	 */
	public User getUser() { return user; }

	/**
	 * @param user the user to set
	 * @since Envoy v0.2-alpha
	 */
	public void setUser(User user) { this.user = user; }

	/**
	 * @return the message ID generator
	 * @since Envoy v0.3-alpha
	 */
	public IdGenerator getIdGenerator() { return idGenerator; }

	/**
	 * @param idGenerator the message ID generator to set
	 * @since Envoy v0.3-alpha
	 */
	public void setIdGenerator(IdGenerator idGenerator) { this.idGenerator = idGenerator; }

	/**
	 * @return {@code true} if an {@link IdGenerator} is present
	 * @since Envoy v0.3-alpha
	 */
	public boolean hasIdGenerator() { return idGenerator != null; }

	/**
	 * @return the offline message cache
	 * @since Envoy v0.3-alpha
	 */
	public Cache<Message> getMessageCache() { return messageCache; }

	/**
	 * @param messageCache the offline message cache to set
	 * @since Envoy v0.3-alpha
	 */
	public void setMessageCache(Cache<Message> messageCache) { this.messageCache = messageCache; }

	/**
	 * @return the offline status cache
	 * @since Envoy v0.3-alpha
	 */
	public Cache<MessageStatusChangeEvent> getStatusCache() { return statusCache; }

	/**
	 * @param statusCache the offline status cache to set
	 * @since Envoy v0.3-alpha
	 */
	public void setStatusCache(Cache<MessageStatusChangeEvent> statusCache) { this.statusCache = statusCache; }

	/**
	 * Searches for a message by ID.
	 * 
	 * @param id the ID of the message to search for
	 * @return the message with the corresponding ID, or {@code null} if no message
	 *         has been found
	 * @since Envoy v0.1-beta
	 */
	public Message getMessage(long id) {
		for (Chat c : chats)
			for (Message m : c.getModel())
				if (m.getId() == id) return m;
		return null;
	}
}
