package envoy.client.ui;

import java.awt.EventQueue;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import envoy.client.data.*;
import envoy.client.net.Client;
import envoy.client.net.WriteProxy;
import envoy.data.Config;
import envoy.data.Message;
import envoy.data.User.UserStatus;
import envoy.exception.EnvoyException;
import envoy.util.EnvoyLog;

/**
 * Starts the Envoy client and prompts the user to enter their name.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>Startup.java</strong><br>
 * Created: <strong>12 Oct 2019</strong><br>
 *
 * @author Leon Hofmeister
 * @author Maximilian K&auml;fer
 * @author Kai S. K. Engelbart
 * @since Envoy v0.1-alpha
 */
public class Startup {

	private static ChatWindow chatWindow;

	private static final Logger logger = EnvoyLog.getLogger(Startup.class);

	/**
	 * Loads the application by first loading the configuration, then acquiring a
	 * user name and connecting to the server. If the server cannot be reached,
	 * offline mode is entered if possible. After that, a {@link ChatWindow}
	 * instance is initialized and then displayed to the user. Upon application
	 * exit, settings and the local database are saved.
	 *
	 * @param args the command line arguments may contain configuration parameters
	 *             and are parsed by the {@link Config} class
	 * @since Envoy v0.1-alpha
	 */
	public static void main(String[] args) {
		ClientConfig config = ClientConfig.getInstance();
		SwingUtilities.invokeLater(() -> chatWindow = new ChatWindow());

		try {
			// Load the configuration from client.properties first
			Properties properties = new Properties();
			properties.load(Startup.class.getClassLoader().getResourceAsStream("client.properties"));
			config.load(properties);

			// Override configuration values with command line arguments
			if (args.length > 0) config.load(args);

			// Check if all mandatory configuration values have been initialized
			if (!config.isInitialized()) throw new EnvoyException("Configuration is not fully initialized");
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null, "Error loading configuration values:\n" + e, "Configuration error", JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
			System.exit(1);
		}

		// Setup logger for the envoy package
		EnvoyLog.initialize(config);
		EnvoyLog.attach("envoy");
		EnvoyLog.setFileLevelBarrier(config.getFileLevelBarrier());
		EnvoyLog.setConsoleLevelBarrier(config.getConsoleLevelBarrier());

		// Initialize the local database
		LocalDb localDb;
		if (config.isIgnoreLocalDB()) {
			localDb = new TransientLocalDb();
			JOptionPane.showMessageDialog(null,
					"Ignoring local database.\nMessages will not be saved!",
					"Local database warning",
					JOptionPane.WARNING_MESSAGE);
		} else try {
			localDb = new PersistentLocalDb(new File(config.getHomeDirectory(), config.getLocalDB().getPath()));
		} catch (IOException e3) {
			logger.log(Level.SEVERE, "Could not initialize local database", e3);
			JOptionPane.showMessageDialog(null, "Could not initialize local database!\n" + e3, "Local database error", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
			return;
		}

		// Initialize client and unread message cache
		Client			client	= new Client();
		Cache<Message>	cache	= new Cache<>();

		// Try to connect to the server
		new LoginDialog(client, localDb, cache);
		SwingUtilities.invokeLater(() -> chatWindow.setVisible(true));

		// Set client user in local database
		localDb.setUser(client.getSender());

		// Initialize chats in local database
		try {
			localDb.initializeUserStorage();
			localDb.loadUserData();
		} catch (FileNotFoundException e) {
			// The local database file has not yet been created, probably first login
		} catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null,
					"Error while loading local database: " + e + "\nChats will not be stored locally.",
					"Local DB error",
					JOptionPane.WARNING_MESSAGE);
		}

		// Initialize write proxy
		final WriteProxy writeProxy = client.createWriteProxy(localDb);

		if (client.isOnline()) {

			// Save all users to the local database and flush cache
			localDb.setUsers(client.getUsers());
			writeProxy.flushCache();
		} else 
			// Set all contacts to offline mode
			localDb.getUsers().values().stream().filter(u -> u != localDb.getUser()).forEach(u -> u.setStatus(UserStatus.OFFLINE));

		// Display ChatWindow and StatusTrayIcon
		EventQueue.invokeLater(() -> {
			try {
				chatWindow.initContent(client, localDb, writeProxy);

				// Relay unread messages from cache
				if (cache != null && client.isOnline()) cache.relay();

				try {
					new StatusTrayIcon(chatWindow).show();

					// If the tray icon is supported and corresponding settings is set, hide the
					// chat window on close
					Settings.getInstance()
						.getItems()
						.get("onCloseMode")
						.setChangeHandler((onCloseMode) -> chatWindow
							.setDefaultCloseOperation((Boolean) onCloseMode ? JFrame.HIDE_ON_CLOSE : JFrame.EXIT_ON_CLOSE));
				} catch (EnvoyException e) {
					logger.warning("The StatusTrayIcon is not supported on this platform!");
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		});

		// Save Settings and PersistentLocalDb on shutdown
		Runtime.getRuntime().addShutdownHook(new Thread(() -> {
			try {
				logger.info("Closing connection...");
				client.close();

				logger.info("Saving local database and settings...");
				localDb.save();
				Settings.getInstance().save();
			} catch (Exception e) {
				logger.log(Level.SEVERE, "Unable to save local files", e);
			}
		}));
	}
}
