package envoy.client.ui.list;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.*;

/**
 * Provides a vertical list layout of components provided in a
 * {@link ComponentListModel}. Similar to {@link javax.swing.JList} but capable
 * of rendering {@link JPanel}s.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>ComponentList.java</strong><br>
 * Created: <strong>25.01.2020</strong><br>
 *
 * @param <E> the type of object displayed in this list
 * @author Kai S. K. Engelbart
 * @since Envoy v0.3-alpha
 */
public class ComponentList<E> extends JPanel {

	private ComponentListModel<E>			model;
	private ComponentListCellRenderer<E>	renderer;

	private int currentSelection = -1;

	private static final long serialVersionUID = 1759644503942876737L;

	/**
	 * Creates an instance of {@link ComponentList}.
	 *
	 * @param renderer the list cell renderer used to display elements provided by
	 *                 the {@link ComponentListModel}
	 * @since Envoy v0.3-alpha
	 */
	public ComponentList(ComponentListCellRenderer<E> renderer) {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		this.renderer = renderer;
	}

	/**
	 * Creates an instance of {@link ComponentList}.
	 *
	 * @param model    the list model providing the list elements to render
	 * @param renderer the list cell renderer used to display elements provided by
	 *                 the {@link ComponentListModel}
	 * @since Envoy v0.3-alpha
	 */
	public ComponentList(ComponentListModel<E> model, ComponentListCellRenderer<E> renderer) {
		this(renderer);
		this.model = model;
		setModel(model);
	}

	/**
	 * Sets the list model providing the list elements to render. The rendered
	 * components will be synchronized with the contents of the new model or removed
	 * if the new model is {@code null}.
	 *
	 * @param model the list model to set
	 * @since Envoy v0.3-alpha
	 */
	public void setModel(ComponentListModel<E> model) {

		// Remove old model
		if (this.model != null) this.model.setComponentList(null);

		// Synchronize with new model
		this.model = model;
		if (model != null) this.model.setComponentList(this);
		synchronizeModel();
	}

	/**
	 * Removes all child components and then adds all components representing the
	 * elements of the {@link ComponentListModel}.
	 *
	 * @since Envoy v0.3-alpha
	 */
	public void synchronizeModel() {
		removeAll();
		if (model != null) model.forEach(this::add);
		revalidate();
	}

	/**
	 * Adds an object to the list by rendering it with the current
	 * {@link ComponentListCellRenderer}.
	 *
	 * @param elem the element to add
	 * @since Envoy v0.3-alpha
	 */
	void add(E elem) { add(elem, getComponentCount(), false); }

	/**
	 * Adds an object to the list by rendering it with the current
	 * {@link ComponentListRenderer}.
	 *
	 * @param elem       the element to add
	 * @param index      the index at which to add the element
	 * @param isSelected the selection state of the element
	 * @since Envoy v0.1-beta
	 */
	private void add(E elem, int index, boolean isSelected) {
		final JComponent component = renderer.getListCellComponent(this, elem, isSelected);
		component.addMouseListener(getSelectionListener(index));
		add(component, index);
	}

	/**
	 * @param componentIndex the index of the list component to which the mouse
	 *                       listener will be added
	 * @return a mouse listener calling the
	 *         {@link ComponentList#componentSelected(int)} method with the
	 *         component's index when a left click is performed by the user
	 * @since Envoy v0.1-beta
	 */
	private MouseListener getSelectionListener(int componentIndex) {
		return new MouseAdapter() {

			@Override
			public void mouseClicked(MouseEvent e) { if (SwingUtilities.isLeftMouseButton(e)) componentSelected(componentIndex); }
		};
	}

	/**
	 * Gets called when a list component has been clicked on by the user. Any
	 * previous selections are then removed and the selected component gets
	 * redrawn.<br>
	 * <br>
	 * If the currently selected component gets selected again, the selection is
	 * removed.
	 *
	 * @param index the index of the selected component
	 * @since Envoy v0.1-beta
	 */
	private void componentSelected(int index) {
		if (index == currentSelection) {

			// Clear selection
			update(currentSelection, false);
			currentSelection = -1;
		} else {

			// Remove old selection
			if (currentSelection >= 0) update(currentSelection, false);

			// Assign new selection
			currentSelection = index;

			// Update current selection
			update(currentSelection, true);
		}

		revalidate();
		repaint();
	}

	/**
	 * Replaces a list element with a newly rendered instance of its contents.
	 *
	 * @param index      the index of the element to update
	 * @param isSelected the selection state passed to the {@link ListCellRenderer}
	 * @since Envoy v0.1-beta
	 */
	private void update(int index, boolean isSelected) {
		remove(index);
		add(model.get(index), index, isSelected);
	}
}
