package envoy.client.ui.renderer;

import java.awt.Font;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.EnumMap;

import javax.imageio.ImageIO;
import javax.swing.*;

import envoy.client.data.Settings;
import envoy.client.ui.Color;
import envoy.client.ui.Theme;
import envoy.client.ui.list.ComponentList;
import envoy.client.ui.list.ComponentListCellRenderer;
import envoy.data.Message;
import envoy.data.Message.MessageStatus;

/**
 * Defines how a message is displayed.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>MessageListRenderer.java</strong><br>
 * Created: <strong>19 Oct 2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @author Leon Hofmeister
 * @since Envoy v0.1-alpha
 */
public class MessageListRenderer implements ComponentListCellRenderer<Message> {

	private static final EnumMap<MessageStatus, BufferedImage> statusIcons = new EnumMap<>(MessageStatus.class);

	static {
		for (MessageStatus ms : MessageStatus.values())
			try {
				statusIcons.put(ms, ImageIO.read(MessageListRenderer.class.getResourceAsStream(ms.toString().toLowerCase() + "_icon.png")));
			} catch (IOException e) {
				e.printStackTrace();
			}
	}

	private JTextArea messageTextArea;

	@Override
	public JPanel getListCellComponent(ComponentList<? extends Message> list, Message value, boolean isSelected) {
		final JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
		final Theme theme = Settings.getInstance().getThemes().get(Settings.getInstance().getCurrentTheme());

		// Panel background
		panel.setBackground(isSelected ? theme.getSelectionColor() : theme.getCellColor());

		// TODO: Handle message attachments

		final String	state	= value.getStatus().toString();
		final String	date	= new SimpleDateFormat("dd.MM.yyyy HH:mm").format(value.getCreationDate());
		final String	text	= value.getText();

		// The Label that displays the creation date of a message
		JLabel dateLabel = new JLabel(date);
		// Set the date color to be the value of DateColorChat
		dateLabel.setForeground(theme.getDateColor());

		panel.add(dateLabel);

		if (value.isForwarded()) try {
			var forwardLabel = new JLabel("Forwarded", new ImageIcon(ClassLoader.getSystemResourceAsStream(null).readAllBytes()),
					SwingConstants.CENTER);
			forwardLabel.setBackground(panel.getBackground());
			forwardLabel.setForeground(Color.lightGray);
			panel.add(forwardLabel);
		} catch (IOException e) {
			e.printStackTrace();
		}

		// The JTextArea that displays the text content of a message and its status
		messageTextArea = new JTextArea(text + System.getProperty("line.separator"));
		messageTextArea.setLineWrap(true);
		messageTextArea.setWrapStyleWord(true);
		messageTextArea.setAlignmentX(0.5f);
		messageTextArea.setForeground(theme.getMessageTextColor());
		messageTextArea.setBackground(panel.getBackground());
		messageTextArea.setEditable(false);

		panel.add(messageTextArea);

		JLabel statusLabel = new JLabel(state);
		statusLabel.setFont(new Font("Arial", Font.BOLD, 14));
		Color statusColor;
		switch (value.getStatus()) {
			case WAITING:
				statusColor = Color.gray;
				break;
			case SENT:
				statusColor = Color.blue;
				break;
			case RECEIVED:
				statusColor = Color.yellow;
				break;
			case READ:
				statusColor = Color.green;
				break;
			default:
				statusColor = theme.getMessageTextColor();
				break;
		}
		statusLabel.setForeground(statusColor);
		statusLabel.setBackground(panel.getBackground());

		panel.add(statusLabel);

		// Define some space to the messages below
		panel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(), BorderFactory.createEtchedBorder()));

		return panel;
	}
}
